package physicssim;
import com.interworldtransport.clados.*;
import java.awt.*;
/**
 * <p>Title: physicssim (can you think of a better name?)</p>
 * <p>Description: A physics simulator using clados</p>
 * <p>Copyright: GNU General Public License.  A full copy of the license
 * can be found bundled with this package.
 * <p>Company: </p>
 * @authors Martin Baker and Alfred Differ
 * @version 1.0
 */

/** shape is used as a base class for specific shapes such as sphere
 * this holds information which is common to all shapes
 *
 * It may be that in future we hold all shapes in a common form made up of cubes
 * or we could have many shapes derived from this, such as,
 * sphere
 * cube
 * general (multiple cubes)
 * infinite plane (for floor)
 *
 * I'm not sure 'shape' is the best name (object is no good as it can be confused with
 * the base class in Java)
 * I think it needs to have both shape and physics information, so that this class, or those
 * derived from it, can calulate:
 * 1) movement
 * 2) intersection with other shapes
 * 3) response to forces
 * 4) response to collision (impulse)
 * 5) other physical properties (friction coeficient, etc)
 * 5) appearance (colour, surface texture, information reqired for rendering)
 */
public class shape {

/** this holds position, orientation, volume and time
 * I'm not quite clear about the terminology here? for me location would imply
 * position without orientation, position would imply location + orientation
 * Or do you think it should be the other way round?
 *
 * scalar = time difference (is this something we can leave set to 0 for now?
 *          if, at a later stage, we have lots of shapes it seems wastefull
 *          to increment a time value in every shape in every frame. This
 *          could be a global value?)
 * vector = position of centre of mass in absolute x,y,z coordinates (implies
 *          that we have an absolute coordinate system).
 * bivector = orientation, initial value =(pi,0,0)
 * trivector = volume (= radius^3 * 4*pi/3 in the case of sphere)
 */
protected Monad location;

/** this holds angular and linier velocity
 *
 * scalar = time difference (is this something we can leave set to 0 for now?)
 * vector = linear velocity
 * bivector = angular velocity
 * trivector = 0 since volume of solid objects does not change
 */
protected Monad velocity;

/** allows us to calculate linier acceleration from net force
 */
protected double mass;

/** allows us to calculate angular acceleration from net torque
 * (not used yet, for first phase of this project)
 */
protected double[] inertiaTensor;

/** constructor */
public shape() {
}

/** called by global class prior to rendering next frame
 * this updates the status (location and velocity) of this shape only
 *
 * acceleration is a transitory value used here, since it is calculated each
 * time it does not need to be held as part of the status of this shape?
 */
public void step(shape[] pOthers, long timeSinceLastStep)
{
  try {
      Monad[] Forces = new Monad[pOthers.length];
      Monad ForceSum = new Monad("ForceSum","Initial","Absolute",3,0,"Zero");
      for (int j=0; j<pOthers.length+1; j++) {
            if (pOthers[j]==this) continue; // prevent interaction with self
            else {
                  //In case we want to draw the forces separately...
                  Forces[j]=pOthers[j].physicalAction(location, velocity, mass, timeSinceLastStep);
                  //What we really need, though, is the net force
                  ForceSum.Add(Forces[j]);
                  //There are exceptions we will have to deal with here
                  //since this can break in a number of ways.
            }
      }
      //At this point we can trust that we have a net force monad telling this
      //shape how it must accelerate.  That means it is time to integrate the
      //kinematic variables.
      //Remember that the scale method actually alters the calling monad.
      Monad acceleration=new Monad("Acceleration", ForceSum);
      acceleration.Scale((double)(timeSinceLastStep)/mass);
      velocity.Add(acceleration);
      Monad tvelocity=new Monad(velocity);      //a temporary copy.  High overhead!
      tvelocity.Scale((double)timeSinceLastStep);
      location.Add(tvelocity);
      tvelocity=null;
      //location and velocity now hold monads representing the state of the shape
      //after the time step.  There might be a technique to do this without the
      //overhead of temporary copies.  I'll think on it some.

      //Now it is time to render this shape.  We know it's kinematic details, so
      //we have all the information we need
  } catch(Exception e) {
    System.out.print("error in shape.step");
  }
}

/** The PhysicalAction method is called by shapes asking for forces from this shape.
 * Respond to requests from other Shapes for a force monad.  This is the force
 *  this particular shape applies to others.  A shape should never call its own
 *  physicalAction method.
 */
public Monad physicalAction(Monad pLoc, Monad pVel, double mass, long pTime) {
      //This version of the method is better set up for a floor-like shape than it
      //is for some other sphere.  With a bit of hard work, though, the force
      //calculation section below could be generalized.

      Monad result =null;
  try {
      Monad tempLoc=new Monad(pLoc);
      tempLoc.GradePart(3);
      double radius=Math.pow(0.75*tempLoc.MagnitudeOf()/Math.PI, 0.333333);

      tempLoc=new Monad(pLoc);
      tempLoc.Subtract(location);   //subtract out location of this shape
      tempLoc.GradePart(1);         //only the vector part will be used to decide collision
      double position=tempLoc.MagnitudeOf();
      tempLoc=null; //clean up

      if (position<=radius) {
            //Calling shape is too close to this shape, so a collision force results
            //Need to make linear result come out as 2*mass*linear velocity/timestep

            double[] tCoord={0, 0, 0, 2.0*mass/pTime, 0, 0, 0, 0, 0};
            result = new Monad("collision", "Initial", "Absolute", 3, 0, tCoord);

            Monad tempVel =new Monad(pVel);
            tempVel.GradePart(1);
            double velMag=tempVel.MagnitudeOf();
            result.Scale(velMag);
      } else {
            //Calling shape is far away, so a gravity force results.

            double[] tCoord={0, 0, 0, -9.81*mass, 0, 0, 0, 0, 0};
            result = new Monad("gravity", "Initial", "Absolute", 3, 0, tCoord);
      }
  } catch(Exception e) {
    System.out.print("error in shape.physicalAction");
  }
  return result;
}


/** test for an intersection between this and other
 * I expect this class to be overriden for each shape
 * for instance sphere would test if distance > x*x + y*y + z*z
 *
 * the problem is other may be a different shape than this, so
 * other may be any shape type
 */
/*public boolean intersection(shape other){
  // for phase 1 assume other is ground plane
  double[] coords=location.getCoeff();
  double radius = coords[8] *3/(Math.PI*4); // should really be cubed root of this ugh!
  if (coords[2]<radius) return true; // is y coordinate less than radius(has it hit ground)
  return false;
}*/

/** overriden by box and sphere
 *
 */
public void paint(Graphics g) {

}
}